import { useEffect, useMemo, useState, useRef } from "react";
import { useFrame } from "@react-three/fiber";
import * as THREE from "three";
import { usePlayButtonStore } from "../../../store/usePlayButtonStore";
import { useSimulationStates } from "../../../store/store";


interface StaticMachineState {
    uuid: string;
    status: string;
    actions: { uuid: string; name: string; buffer: number; material: string; };
    machineTriggerId: string;
    connectedArmBot: string;
}

interface ArmBotState {
    uuid: string;
    position: [number, number, number];
    rotation: [number, number, number];
    status: string;
    material: string;
    triggerId: string;
    connections: {
        source: { modelUUID: string; pointUUID: string };
        targets: { modelUUID: string; pointUUID: string }[];
    };
    actions: { uuid: string; name: string; speed: number; processes: { triggerId: string; startPoint: string; endPoint: string }[]; };
    isActive?: boolean;
}

type IKAnimationControllerProps = {
    ikSolver: any;
    process: {
        triggerId: string;
        startPoint: THREE.Vector3;
        endPoint: THREE.Vector3;
        speed: number;
    }[];
    selectedTrigger: string;
    targetBoneName: string;
    uuid: string;
    logStatus: (status: string) => void;
    groupRef: React.RefObject<THREE.Group>;
    armBot: ArmBotState;
    setArmBots: React.Dispatch<React.SetStateAction<ArmBotState[]>>;
    setStaticMachines: React.Dispatch<React.SetStateAction<StaticMachineState[]>>;
    updateArmBotStatus: (status: string) => void;
}

const IKAnimationController = ({
    ikSolver,
    process,
    selectedTrigger,
    targetBoneName,
    uuid,
    logStatus,
    groupRef,
    armBot,
    setArmBots,
    setStaticMachines,
    updateArmBotStatus
}: IKAnimationControllerProps) => {
    const [progress, setProgress] = useState(0);
    const [initialProgress, setInitialProgress] = useState(0);
    const [needsInitialMovement, setNeedsInitialMovement] = useState(true);
    const [isInitializing, setIsInitializing] = useState(true);
    const restSpeed = 0.1;
    const restPosition = new THREE.Vector3(0, 2, 1.6);
    const { isPlaying } = usePlayButtonStore();
    const { simulationStates } = useSimulationStates();

    // Track previous states for comparison
    const prevStateRef = useRef({
        isInitializing: true,
        needsInitialMovement: true,
        selectedTrigger: "",
        progress: 0
    });

    // Track previous status for comparison
    const prevStatusRef = useRef("");

    const initialCurveRef = useRef<THREE.CatmullRomCurve3 | null>(null);
    const initialStartPositionRef = useRef<THREE.Vector3 | null>(null);

    useEffect(() => {
        setProgress(0);
    }, [selectedTrigger]);

    useEffect(() => {
        if (ikSolver) {
            const targetBone = ikSolver.mesh.skeleton.bones.find(
                (b: any) => b.name === targetBoneName
            );
            if (targetBone) {
                initialStartPositionRef.current = targetBone.position.clone();
                calculateInitialCurve(targetBone.position);
                logStatus(`[Arm ${uuid}] Initializing IK system, starting position: ${targetBone.position.toArray()}`);
            }
        }
    }, [ikSolver]);

    // Log state changes
    useEffect(() => {
        const prev = prevStateRef.current;

        if (prev.isInitializing !== isInitializing) {
            if (!isInitializing) {
                logStatus(`[Arm ${uuid}] Completed initialization, now at rest position`);
            }
        }

        if (prev.needsInitialMovement !== needsInitialMovement && !needsInitialMovement) {
            logStatus(`[Arm ${uuid}] Reached rest position, ready for animation`);

        }

        if (prev.selectedTrigger !== selectedTrigger) {
            logStatus(`[Arm ${uuid}] Processing new trigger: ${selectedTrigger}`);

            const currentProcess = process.find(p => p.triggerId === prev.selectedTrigger);
            if (currentProcess) {
                const triggerId = currentProcess.triggerId;

                const endPoint = armBot.actions.processes.find((process) => process.triggerId === triggerId)?.endPoint;

                // Search simulationStates for a StaticMachine or Conveyor that has a point matching this endPointId
                const matchedMachine = simulationStates.find((state) => {
                    if (state.type === "Conveyor") {
                        // For Conveyor, points is an array
                        return (state).points.some(
                            (point) => point.uuid === endPoint
                        );
                    } else if (state.type === "StaticMachine") {
                        // For StaticMachine, points is an object
                        return state.points.uuid === endPoint;
                    }
                    return false;
                });

                if (matchedMachine) {
                    // Log if the end point is a conveyor
                    if (matchedMachine.type === "Conveyor") {
                        logStatus(`[Arm ${uuid}] Reached end point which is a conveyor (${matchedMachine.modelName})`);
                    } else {
                        logStatus(`[Arm ${uuid}] Reached end point which is a static machine (${matchedMachine.modelName})`);
                    }

                    if (matchedMachine.type === "StaticMachine") {
                        setStaticMachines((machines) => {
                            return machines.map((machine) => {
                                if (machine.uuid === matchedMachine.modeluuid) {
                                    return { ...machine, status: "running" };
                                } else {
                                    return machine;
                                }
                            });
                        });
                    }

                    if (matchedMachine.type === "Conveyor") {
                        setArmBots((prev) =>
                            prev.map((arm) => {
                                if (arm.uuid === uuid) {
                                    return {
                                        ...arm,
                                        isActive: false
                                    };
                                }
                                else {
                                    return arm;
                                }
                            })
                        );
                    }
                }
            }
        }

        // Update previous state
        prevStateRef.current = {
            isInitializing,
            needsInitialMovement,
            selectedTrigger,
            progress
        };
    }, [isInitializing, needsInitialMovement, selectedTrigger, progress]);

    const calculateInitialCurve = (startPosition: THREE.Vector3) => {
        const direction = new THREE.Vector3().subVectors(restPosition, startPosition);
        const distance = direction.length();
        direction.normalize();

        const perpendicular = new THREE.Vector3(-direction.z, 0, direction.x).normalize();

        const midHeight = 0.5;
        const tiltAmount = 1;
        const mid = new THREE.Vector3()
            .addVectors(startPosition, restPosition)
            .multiplyScalar(0.5)
            .add(perpendicular.clone().multiplyScalar(distance * 0.3 * tiltAmount))
            .add(new THREE.Vector3(0, midHeight, 0));

        initialCurveRef.current = new THREE.CatmullRomCurve3([
            startPosition,
            new THREE.Vector3().lerpVectors(startPosition, mid, 0.33),
            mid,
            new THREE.Vector3().lerpVectors(mid, restPosition, 0.66),
            restPosition
        ]);
    };

    const processedCurves = useMemo(() => {
        if (isPlaying)
            return process.map((p) => {
                const tempLift = 0.5;
                const localStart = groupRef.current?.worldToLocal(p.startPoint.clone().add(new THREE.Vector3(0, tempLift, 0)));
                const localEnd = groupRef.current?.worldToLocal(p.endPoint.clone().add(new THREE.Vector3(0, tempLift, 0)));

                if (localStart && localEnd) {

                    const mid = new THREE.Vector3(
                        (localStart.x + localEnd.x) / 1,
                        Math.max(localStart.y, localEnd.y) + 0.8,
                        (localStart.z + localEnd.z) / 0.9
                    );

                    const points = [
                        restPosition.clone(),
                        localStart.clone(),
                        mid.clone(),
                        localEnd.clone(),
                        restPosition.clone(),
                    ];
                    const curve = new THREE.CatmullRomCurve3(points);
                    const restToStartDist = points[0].distanceTo(points[1]);
                    const startToEndDist = points[1].distanceTo(points[3]);
                    const endToRestDist = points[3].distanceTo(points[4]);

                    const totalDist = restToStartDist + startToEndDist + endToRestDist;
                    const restToStartRange = [0, restToStartDist / totalDist];
                    const startToEndRange = [
                        restToStartRange[1],
                        restToStartRange[1] + startToEndDist / totalDist,
                    ];
                    const endToRestRange = [startToEndRange[1], 1];

                    return {
                        trigger: p.triggerId,
                        curve,
                        speed: p.speed,
                        restToStartRange,
                        startToEndRange,
                        endToRestRange,
                    };
                }
            });
    }, [process, groupRef, isPlaying]);

    const activeCurve = useMemo(() => {
        if (isPlaying && processedCurves)
            return processedCurves.find((c) => c?.trigger === selectedTrigger);
    }, [processedCurves, selectedTrigger, isPlaying]);

    // Initial movement to rest position
    useFrame((_, delta) => {
        if (!ikSolver || !needsInitialMovement || !isInitializing || !initialCurveRef.current) return;

        const targetBone = ikSolver.mesh.skeleton.bones.find(
            (b: any) => b.name === targetBoneName
        );
        if (!targetBone) return;

        setInitialProgress((prev) => {
            const next = prev + delta * 0.5;
            if (next >= 1) {
                targetBone.position.copy(restPosition);
                setNeedsInitialMovement(false);
                setIsInitializing(false);
                return 1;
            }
            targetBone.position.copy(initialCurveRef.current!.getPoint(next));
            return next;
        });

        ikSolver.update();
    });

    // Main animation loop
    useFrame((_, delta) => {
        if (!ikSolver || !activeCurve || isInitializing || !isPlaying) return;

        const { curve, speed, restToStartRange, startToEndRange, endToRestRange } = activeCurve;
        const targetBone = ikSolver.mesh.skeleton.bones.find(
            (b: any) => b.name === targetBoneName
        );
        if (!targetBone) return;

        let currentSpeed = restSpeed;
        let currentStatus = "idle"; // Default status

        // Determine current phase and status
        if (progress < restToStartRange[1]) {
            currentSpeed = restSpeed;
            currentStatus = "moving"; // Moving to start point
        } else if (progress >= startToEndRange[0] && progress < startToEndRange[1]) {
            currentSpeed = speed;
            currentStatus = "moving"; // Moving between points
        } else if (progress >= endToRestRange[0] && progress < 1) {
            currentSpeed = restSpeed;
            currentStatus = "moving"; // Returning to rest
        } else if (progress >= 1) {
            currentStatus = "idle"; // Completed cycle
        }

        // Update status when it changes
        if (prevStatusRef.current !== currentStatus) {
            updateArmBotStatus(currentStatus);
            prevStatusRef.current = currentStatus;
        }

        // Only update progress if we're not already at the end
        if (progress < 1) {
            setProgress((prev) => {
                const next = prev + delta * currentSpeed;
                return Math.min(next, 1); // Cap at 1
            });
        }

        // Update bone position based on progress
        if (progress < 1) {
            targetBone.position.copy(curve.getPoint(progress));
        } else {
            targetBone.position.copy(curve.getPoint(1));
        }

        ikSolver.update();
    });

    return null;
};

export default IKAnimationController;