import * as THREE from "three";
import { useEffect, useMemo, useRef, useState } from "react";
import { useFrame, useLoader } from "@react-three/fiber";
import { DRACOLoader } from "three/examples/jsm/loaders/DRACOLoader";
import { GLTFLoader } from "three/examples/jsm/loaders/GLTFLoader";
import { clone } from "three/examples/jsm/utils/SkeletonUtils";
import { CCDIKSolver, CCDIKHelper, } from "three/examples/jsm/animation/CCDIKSolver";
import IKAnimationController from "./IKAnimationController";
import { TransformControls } from "@react-three/drei";

interface StaticMachineState {
    uuid: string;
    status: string;
    actions: { uuid: string; name: string; buffer: number; material: string; };
    machineTriggerId: string;
    connectedArmBot: string;
}

interface ArmBotState {
    uuid: string;
    position: [number, number, number];
    rotation: [number, number, number];
    status: string;
    material: string;
    triggerId: string;
    connections: {
        source: { modelUUID: string; pointUUID: string };
        targets: { modelUUID: string; pointUUID: string }[];
    };
    actions: { uuid: string; name: string; speed: number; processes: { triggerId: string; startPoint: string; endPoint: string }[]; };
    isActive?: boolean;
}

const IkInstances = ({
    uuid,
    selectedTrigger,
    modelUrl,
    processes,
    position,
    rotation,
    armBot,
    setArmBots,
    setStaticMachines,
    updateArmBotStatus
}: {
    uuid: string;
    selectedTrigger: string;
    modelUrl: string;
    processes: any;
    position: [number, number, number];
    rotation: [number, number, number];
    armBot: ArmBotState;
    setArmBots: React.Dispatch<React.SetStateAction<ArmBotState[]>>;
    setStaticMachines: React.Dispatch<React.SetStateAction<StaticMachineState[]>>;
    updateArmBotStatus: (status: string) => void;
}) => {
    const [ikSolver, setIkSolver] = useState<any>(null);
    const gltf = useLoader(GLTFLoader, modelUrl, (loader) => {
        const draco = new DRACOLoader();
        draco.setDecoderPath("https://cdn.jsdelivr.net/npm/three@0.160.0/examples/jsm/libs/draco/");
        loader.setDRACOLoader(draco);
    });
    const cloned = useMemo(() => clone(gltf.scene), [gltf]);
    const groupRef = useRef<any>(null);
    const targetBoneName = "Target";
    const skinnedMeshName = "link_0";

    useEffect(() => {
        if (!gltf) return;
        const OOI: any = {};
        cloned.traverse((n: any) => {
            if (n.name === targetBoneName) OOI.Target_Bone = n;
            if (n.name === skinnedMeshName) OOI.Skinned_Mesh = n;
        });

        if (!OOI.Target_Bone || !OOI.Skinned_Mesh) return;

        const iks = [
            {
                target: 7,
                effector: 6,
                links: [
                    {
                        index: 5,
                        enabled: true,
                        rotationMin: new THREE.Vector3(-Math.PI / 2, 0, 0),
                        rotationMax: new THREE.Vector3(Math.PI / 2, 0, 0),
                    },
                    {
                        index: 4,
                        enabled: true,
                        rotationMin: new THREE.Vector3(-Math.PI / 2, 0, 0),
                        rotationMax: new THREE.Vector3(0, 0, 0),
                    },
                    {
                        index: 3,
                        enabled: true,
                        rotationMin: new THREE.Vector3(0, 0, 0),
                        rotationMax: new THREE.Vector3(2, 0, 0),
                    },
                    { index: 1, enabled: true, limitation: new THREE.Vector3(0, 1, 0) },
                    { index: 0, enabled: false, limitation: new THREE.Vector3(0, 0, 0) },
                ],
            },
        ];

        const solver = new CCDIKSolver(OOI.Skinned_Mesh, iks);
        setIkSolver(solver);

        const helper = new CCDIKHelper(OOI.Skinned_Mesh, iks, 0.05);
        // groupRef.current.add(helper);

    }, [gltf]);


    const logStatus = (status: string) => {
        // console.log(status);
    }

    return (
        <>
            <group
                ref={groupRef}
                position={position}
                rotation={rotation}
            >
                <primitive
                    object={cloned}
                    scale={[1, 1, 1]}
                    name={`arm-bot`}
                />
            </group>
            <IKAnimationController
                ikSolver={ikSolver}
                process={processes}
                selectedTrigger={selectedTrigger}
                targetBoneName={targetBoneName}
                uuid={uuid}
                logStatus={logStatus}
                groupRef={groupRef}
                armBot={armBot}
                setArmBots={setArmBots}
                setStaticMachines={setStaticMachines}
                updateArmBotStatus={updateArmBotStatus}
            />
        </>
    );
};

export default IkInstances;